/*
** Module   : WindowTitleBar.java
** Abstract : Implementation of a GUI window title bar.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 MAG 20140902 Use GUI driver system color API to get default colors for widgets drawing.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20140929 Make class public.
** 006 CA  20141021 Finished support for the caption buttons, drawing, and associated system menu.
** 007 MAG 20140929 Eliminated call to AbstarctContainer.draw.
** 008 HC  20141021 Implemented window move/position support.
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 CA  20141123 Fixed drawing of window title bar, to reflect if the window is the one with
**                  current focus or not.  Changes in drawing of the caption buttons/icon/title to
**                  reflect the OS drawing.
** 011 EVL 20141204 Adding title bar icon changing capability.
** 012 EVL 20150108 Taking into loaded icon size difference from either LOAD-SMALL-ICON or
**                  LOAD-ICON sources.  Fix the default application icon to be set as gclogo.ico
**                  embedded into runtime jar file.  Rework for icon/image implementation.
*/

package com.goldencode.p2j.ui.client.gui;

import java.io.*;
import java.awt.Font;
import java.awt.event.*;
import java.util.*;
import java.util.logging.*;

import javax.imageio.*;
import javax.imageio.stream.*;

import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Implementation of a GUI window title bar.  This consists of (from left to right):
 * <ul>
 *    <li>an {@link #icon}</li>
 *    <li>a {@link #title}</li>
 *    <li>caption {@link #buttons}</li>
 * </ul>
 * Each element is a specific widget instance.
 */
public class WindowTitleBar
extends AbstractContainer<GuiOutputManager>
implements MouseListener,
           MouseMotionListener
{
   /** The window's icon. */
   WindowIcon icon;
   
   /** The window title. */
   WindowTitle title;

   /** The graphics driver. */
   private GuiDriver gd;

   /** The caption buttons. */
   private CaptionButton[] buttons;

   /** This widget configuration instance. */
   private BaseConfig config;
   
   /**
    * Basic c'tor.
    * 
    * @param    id
    *           The widget's ID.
    */
   WindowTitleBar(WidgetId id)
   {
      nativeInsets = new NativeInsets(2, 2, 2, 2);

      gd = (GuiDriver) OutputManager.getDriver();
      
      config = ConfigManager.getInstance().resolveWidgetConfig(id, BaseConfig.class);
      
      icon = new WindowIcon(WidgetId.nextID());
      byte[] appIcon = Window.getWindowDefaultIcon();
      // found the application icon defined on the server side
      if (appIcon != null)
      {
         // at this time we have either application icon or default icon loaded
         icon.setIconData(appIcon);
      }
         
      this.add(icon);

      title = new WindowTitle(WidgetId.nextID());
      this.add(title);

      CaptionButtonType[] types =
      {
         CaptionButtonType.ICONIFY,
         CaptionButtonType.MAXIMIZE,
         CaptionButtonType.CLOSE
      };
      buttons = new CaptionButton[types.length];
      for (int i = 0; i < types.length; i++)
      {
         CaptionButton btn = new CaptionButton(WidgetId.nextID(), types[i]);
         this.add(btn);

         buttons[i] = btn;
      }
   }
   
   /**
    * Get the widget config.
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   @Override
   public BaseConfig config()
   {
      return config;
   }

   /**
    * Layout the elements of this title.   Everything is done in pixel units.
    */
   public void doLayout()
   {
      WindowGuiImpl window = (WindowGuiImpl) ancestor();
      
      CoordinatesConversion c = window.screen().coordinates();

      int font = window.resolveTitleFont();
      // compute the window title's font height, in pixels.  this will determine the titlebar's
      // height.
      int fheight = c.heightToNative(1, window.config().id, font) + 
                    nativeInsets.top + 
                    nativeInsets.bottom;

      // - icon on left-side; to ensure the character unit location is properly in sync with the 
      // native location, compute the character location first and re-transform to native for the
      // physical location.
      icon.setLocation(c.widthFromNative(nativeInsets.left), c.heightFromNative(nativeInsets.top));
      NativePoint iconLoc = c.pointToNative(icon.location());
      icon.setPhysicalLocation(iconLoc.x, iconLoc.y);

      // title height can not be less than the defualt caption button size
      fheight = Math.max(fheight, GuiOptions.CAPTION_BUTTON_SIZE);
      
      // title height: needs to be done early, icon's height depends on it
      title.setHeight(c.heightFromNative(fheight));

      // entire bar height is compute in native units, then transformed to character units.
      // this is needed because when transforming metrics between native and character units,
      // these functions are not associative (i.e. T(a+b+c) does not necessarily give the same
      // result as T(a)+T(b)+T(c), where T is a transformation function).
      double barHeight = c.heightFromNative(c.heightToNative(title.height()) + 
                                            nativeInsets.top                 + 
                                            nativeInsets.bottom);
      setHeight(barHeight);
      
      // - minimize, maximize, close buttons
   
      int ndim = GuiOptions.CAPTION_BUTTON_SIZE;
      double bw = c.widthFromNative(ndim);
      double bh = c.heightFromNative(ndim);
      
      // add the buttons - all coordinates are relative to this container.
      double bx = this.width();
      // determine the vertical location so that the button is centered vertically
      double by = c.heightFromNative((c.heightToNative(barHeight) - ndim) / 2);

      for (int i = buttons.length - 1; i >= 0; i--)
      {
         CaptionButton btn  = buttons[i];
         btn.setDimension(bw, bh);
         
         bx = bx - c.widthFromNative(c.widthToNative(bw) + nativeInsets.right);
         btn.setLocation(bx, by);

         NativePoint btnLoc = c.pointToNative(btn.location());
         btn.setPhysicalLocation(btnLoc.x, btnLoc.y);
      }

      // - title in the middle, between the icon and the minimize button
      title.setLocation(icon.bounds().right() + 
                        c.widthFromNative(nativeInsets.right + nativeInsets.left),
                        c.heightFromNative(nativeInsets.top));
      NativePoint titleLoc = c.pointToNative(title.location());
      title.setPhysicalLocation(titleLoc.x, titleLoc.y);

      // set title width
      double twidth = buttons[0].bounds().left() - 
                      title.location().x         - 
                      c.widthFromNative(nativeInsets.right);
      title.setWidth(twidth);
   }
   
   /**
    * Get widget minimum size - same as {@link #dimension()}.
    *
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }

   /**
    * Move this widget on the X and Y axis, by the specified deltas.  Values are specified in
    * units depending on the UI type.
    * 
    * @param   dx
    *          The X-axis delta.
    * @param   dy
    *          The Y-axis delta.
    */
   @Override
   public void moveBy(int dx, int dy)
   {
      window().moveBy(dx, dy);
   }

   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(NativePoint p)
   {
      Widget<GuiOutputManager> w = super.findMouseSource(p);
      
      return (w != null ? w : this);
   }
   
   /**
    * Draw the title's background.
    */
   @Override
   public void draw()
   {
      // drawing is done only once, after each (re)layout
      WindowTitleBar bar = WindowTitleBar.this;
      NativePoint origin = bar.physicalLocation();
      final NativeDimension dim = bar.physicalDimension();
      final NativeRectangle clip = clipRectangle(origin, dim);
      
      final WindowGuiImpl window = (WindowGuiImpl) ancestor();
      
      // fill title bar
      gd.draw(origin, clip, new Runnable()
      {
         @Override
         public void run()
         {
            int color = window.isTopWindow() 
                           ? SysColor.COLOR_ACTIVECAPTION
                           : SysColor.COLOR_INACTIVECAPTION;
            
            gd.setColor(gd.getSysColor(color));
            gd.fill3DRect(0, 0, dim.width, dim.height, true);

            // draw contained widgets
            for (Widget<?> widget : widgets())
            {
               widget.draw();
            }
         }
      });
   }
   
   /**
    * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public boolean raiseLegacyMouseEvents()
   {
      return false;
   }

   /**
    * The window's icon.
    */
   @SuppressWarnings("unchecked")
   class WindowIcon
   extends ImageGuiImpl
   implements MouseListener
   {
      /** Flag indicating the icon was loaded by LOAD-SMALL-ICON(). */
      private boolean useSmallIconSize = false;

      /**
       * Create the window icon.
       * 
       * @param    id
       *           The widget's ID.
       */
      public WindowIcon(WidgetId id)
      {
         super(id, true);
      }
      
      /**
       * Get widget height.
       *
       * @return  widget height.
       */
      @Override
      public double height()
      {
         return title.height();
      }

      /**
       * Get widget width.
       *
       * @return  widget width.
       */
      @Override
      public double width()
      {
         WindowGuiImpl window = (WindowGuiImpl) ancestor();
         
         CoordinatesConversion c = window.screen().coordinates();

         return c.widthFromNative(c.heightToNative(height()));
      }

      /**
       * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
       * 
       * @return   Always <code>false</code>.
       */
      @Override
      public boolean raiseLegacyMouseEvents()
      {
         return false;
      }
      
      /**
       * Sets or replaces the window icon data with provided array data.
       * 
       * @param    newIcon
       *           A new icon image array data.
       *
       * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
       */
      public boolean setIconData(byte[] newIcon)
      {
         // the new image candidate
         java.awt.Image imgNew = null;
         
         try
         {
            ImageInputStream iis = ImageIO.createImageInputStream(
                                      new ByteArrayInputStream(newIcon));
            // Get the reader
            Iterator<ImageReader> readers = ImageIO.getImageReaders(iis);

            if (readers.hasNext())
            {
               // we have the reader
               ImageReader reader = readers.next();

               // connect reader to input stream
               reader.setInput(iis);
               // get the icon numbers in a stream allowing internal scanning
               int numIcons = reader.getNumImages(true);
               // loook at approptiate icon size
               for (int i = 0; i < numIcons; i++)
               {
                  int iconSize = Math.max(reader.getWidth(i), reader.getHeight(i)); 
                  if ((useSmallIconSize && iconSize == WindowConfig.SMALL_ICON_SIZE) ||
                      (!useSmallIconSize && iconSize == WindowConfig.ICON_SIZE))
                  {
                     // found the required image, read it and exit icon searching loop
                     imgNew = reader.read(i);
                     break;
                  }
               }
            }

            // have not found required icon size or not icon, use first available image 
            if (imgNew == null)
            {
               imgNew = ImageIO.read(new ByteArrayInputStream(newIcon));
            }
         }
         catch (IOException e)
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.log(Level.SEVERE, "Could not set icon from array...", e);
            }

            imgNew = null;
         }
         
         // image loaded, replace current with new one 
         if (imgNew != null)
         {
            img = imgNew;
            // repaint changed icon
            ThinClient.getInstance().eventDrawingBracket(this, false, true, new Runnable()
            {
               @Override
               public void run()
               {
                  repaint();
               }
            });
            return true;
         }
         else
         {
            // if error while data processing - do nothing and return false
            return false;
         }
      }

      /**
       * Loads the new application icon for given window from the client side filesystem.
       * 
       * @param    small
       *           Flag indicating usage of the small icon case(16x16 in size).
       * @param    name
       *           The fully qualified file name for the icon to be loaded.
       *
       * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
       */
      public boolean loadIcon(boolean small, String name)
      {
         InputStream is = null;
         byte[] icon = null;
         
         // find the icon on local file system
         try
         {
            is = new FileInputStream(name);
         }
         catch (FileNotFoundException ex)
         {
            is = null;
         }
         
         // the file is found, try to read
         if (is != null)
         {
            icon = UiUtils.getData(is);
         }
         
         // if having new icon data - set it
         if (icon != null)
         {
            useSmallIconSize = small;
            return setIconData(icon);
         }
         else
         {
            return false;
         }
      }
   }
   
   /**
    * The window's title.
    */
   @SuppressWarnings("unchecked")
   private class WindowTitle
   extends Label<GuiOutputManager>
   {
      /**
       * Basic c'tor.
       * 
       * @param    id
       *           The widget's ID.
       */
      public WindowTitle(WidgetId id)
      {
         super(id, "");
      }
      
      /**
       * Draw this window title.
       */
      @Override
      public void draw()
      {
         NativePoint origin = physicalLocation();
         final NativeDimension d = physicalDimension();
         final NativeRectangle clip = clipRectangle(origin, d);
         
         gd.draw(origin, clip, new Runnable()
         {
            @Override
            public void run()
            {
               WindowGuiImpl window = (WindowGuiImpl) ancestor();
               
               int color = window.isTopWindow()
                              ? SysColor.COLOR_ACTIVECAPTION
                              : SysColor.COLOR_INACTIVECAPTION;

               // remove old title
               gd.setColor(gd.getSysColor(color));
               gd.fillRect(0, 0, d.width, d.height);

               String title = window.config().title;
               if (title.length() > 0)
               {
                  int titleFont = window.resolveTitleFont();
                  ScreenDriver<?> driver = OutputManager.getDriver();
                  Font font = FontManager.resolveFont(window, titleFont, driver);

                  // draw text
                  gd.setColor(gd.getSysColor(SysColor.COLOR_CAPTIONTEXT));
                  gd.setGuiFont(font);
                  gd.setFontStyle(font.getStyle());

                  int height = gd.getTextHeight(title, font);
                  
                  int y = d.height / 2 + height / 4;
                  int x = 0;
            
                  gd.drawString(title, x, y);
               }
            }
         });
      }

      /**
       * Draw caret - this is a no-op.
       */
      @Override
      public void drawCaret()
      {
         // no-op
      }
      
      /**
       * Find the widget to which the given mouse event can be applied.
       * 
       * @param    p
       *           The mouse physical location.
       * 
       * @return   A widget over which the mouse is positioned or the associated 
       *           {@link WindowTitleBar} if none found.
       */
      @Override
      public Widget<GuiOutputManager> findMouseSource(NativePoint p)
      {
         Widget<GuiOutputManager> w = super.findMouseSource(p);
         
         return (w == null ? WindowTitleBar.this : w);
      }
      
      /**
       * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
       * 
       * @return   Always <code>false</code>.
       */
      @Override
      public boolean raiseLegacyMouseEvents()
      {
         return false;
      }
   }
}
